; -----------------------------------------------------------------------------
;	PSG BGM DRIVER
; -----------------------------------------------------------------------------
;	Copyright (c) 2020 Takayuki Hara
;	http://hraroom.s602.xrea.com/msx/software/index.html
;	
;	Permission is hereby granted, free of charge, to any person obtaining a 
;	copy of this software and associated documentation files (the 
;	"Software"), to deal in the Software without restriction, including 
;	without limitation the rights to use, copy, modify, merge, publish, 
;	distribute, sublicense, and/or sell copies of the Software, and to 
;	permit persons to whom the Software is furnished to do so, subject to 
;	the following conditions:
;	
;	The above copyright notice and this permission notice shall be 
;	included in all copies or substantial portions of the Software.
;	
;	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
;	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
;	MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
;	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
;	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
;	OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION 
;	WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
; =============================================================================
;	2009/09/30	t.hara
;	2019/07/15	t.hara	Modified for ZMA
; -----------------------------------------------------------------------------

include "bgmdriver_d.asm"

; -----------------------------------------------------------------------------
;	AY-3-8910 [PSG] WX^`
; -----------------------------------------------------------------------------
PSG_REG_ADR			= 0xA0		; AhXb`
PSG_REG_WRT			= 0xA1		; f[^Cg

; -----------------------------------------------------------------------------
;	BIOS֘A
; -----------------------------------------------------------------------------
BASE_SLOT			= 0xA8		; {Xbgw|[g
EXT_SLOT			= 0xFFFF	; gXbgw|[g

; -----------------------------------------------------------------------------
;	t^XN\̃ItZbg`
;	INFO_*
; -----------------------------------------------------------------------------
INFO_PLAY_ADR_L		= 0			; tAhX [0Ȃ~]
INFO_PLAY_ADR_H		= 1			; V
INFO_WAIT_COUNT_L	= 2			; ҋ@
INFO_WAIT_COUNT_H	= 3			; V
INFO_EFF_FREQ_L		= 4			; g
INFO_EFF_FREQ_H		= 5			; V
INFO_TONE_FREQ_L	= 6			; ݒg
INFO_TONE_FREQ_H	= 7			; V
INFO_EFF_VOL		= 8			; 
INFO_TONE_VOL		= 9			; ݒ艹
INFO_VIB_WAIT		= 10		; ru[gx
INFO_ENV_STATE		= 11		; Gx[vXe[g
INFO_ENV_VOL		= 12		; Gx[v
INFO_VIB_INDEX		= 13		; ru[gCfbNX
INFO_SOUND_FONT_L	= 14		; Đ̉Ff[^̃AhX
INFO_SOUND_FONT_H	= 15		; V
INFO_NOISE_FREQ		= 16		; mCYg
INFO_SEL_SFONT_L	= 17		; Ff[^̃AhX
INFO_SEL_SFONT_H	= 18		; Ff[^̃AhX
INFO_SIZE			= 19		; INFO\̂̃TCY

; -----------------------------------------------------------------------------
;	Ff[^\̃ItZbg`
;	SFONT_*
; -----------------------------------------------------------------------------
SFONT_VIB_WAVE		= 0			; ru[gg`
SFONT_AR			= 32		; Gx[v AR
SFONT_DR			= 33		; Gx[v DR
SFONT_SL			= 34		; Gx[v SL
SFONT_SR			= 35		; Gx[v SR
SFONT_RR			= 36		; Gx[v RR
SFONT_VIB_WAIT		= 37		; ru[gx
SFONT_NOISE			= 38		; mCYg [SFONT_VIB_WAIT + 1 ł邱ƂO]
SFONT_FREQ_L		= 39		; hpg[L]
SFONT_FREQ_H		= 40		; hpg[H]

; -----------------------------------------------------------------------------
;	
;	input:
;		Ȃ
;	output
;		Ȃ
;	break
;		a, b, c, d, e, f, h, l
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_initialize::
		ld		hl, play_info_ch0
		ld		de, play_info_ch0 + 1
		ld		bc, INFO_SIZE * 3 - 1
		xor		a, a
		ld		[hl], a
		ldir
		ret

; -----------------------------------------------------------------------------
;	tJn
;	input:
;		hl	...	BGMf[^̃AhX
;	output
;		Ȃ
;	break
;		a, b, c, d, e, f, h, l, ix
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_play::
		; ܂̉t~
		push	hl
		call	bgmdriver_stop
		; tJn̂߂̏{
		pop		hl
		di
		; BGMf[^̃AhXۑ
		ld		[play_bgm_data_adr], hl
		ld		ix, [play_bgm_data_adr]
		; ch0 ̉tf[^AhX擾
		ld		e, [ix + 0]
		ld		d, [ix + 1]
		add		hl, de
		ld		[play_info_ch0 + INFO_PLAY_ADR_L], hl
		; ch1 ̉tf[^AhX擾
		ld		hl, [play_bgm_data_adr]
		ld		e, [ix + 2]
		ld		d, [ix + 3]
		add		hl, de
		ld		[play_info_ch1 + INFO_PLAY_ADR_L], hl
		; ch2 ̉tf[^AhX擾
		ld		hl, [play_bgm_data_adr]
		ld		e, [ix + 4]
		ld		d, [ix + 5]
		add		hl, de
		ld		[play_info_ch2 + INFO_PLAY_ADR_L], hl
		; tF[hAEgȂtF[hAEg~
		xor		a, a
		ld		[play_master_volume_speed], a
		ld		[play_master_volume_wait], a
		ld		[play_master_volume], a
		ei
		ret

; -----------------------------------------------------------------------------
;	t~
;	input:
;		Ȃ
;	output
;		Ȃ
;	break
;		a, b, c, d, e, f, h, l
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_stop::
		; ~̓rŊ荞܂ċsRɂȂȂ悤Ɋ
		di
		; ~
		ld		hl, play_info_ch0
		call	bgmdriver_init_play_info
		ld		hl, play_info_ch1
		call	bgmdriver_init_play_info
		ld		hl, play_info_ch2
		call	bgmdriver_init_play_info
		; ։
		ei
		ret

; -----------------------------------------------------------------------------
;	t`FbN
;	input:
;		Ȃ
;	output
;		ZtO ... 1 Ȃ~, 0 Ȃ牉t
;	break
;		a, f, ix
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_check_playing::
		di
		ld		ix, play_info_ch0
		ld		a, [ix + INFO_PLAY_ADR_L]
		or		a, [ix + INFO_PLAY_ADR_H]
		ld		ix, play_info_ch1
		or		a, [ix + INFO_PLAY_ADR_L]
		or		a, [ix + INFO_PLAY_ADR_H]
		ld		ix, play_info_ch2
		or		a, [ix + INFO_PLAY_ADR_L]
		or		a, [ix + INFO_PLAY_ADR_H]
		ei
		ret

; -----------------------------------------------------------------------------
;	tF[hAEg
;	input:
;		a	...	tF[hAEgx[1`255]
;	output
;		Ȃ
;	break
;		a
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_fadeout::
		; ~̓rŊ荞܂ċsRɂȂȂ悤Ɋ
		di
		ld		[play_master_volume_speed], a
		xor		a, a
		ld		[play_master_volume_wait], a
		ld		[play_master_volume], a
		ei
		ret

; -----------------------------------------------------------------------------
;	~
;	input:
;		Ȃ
;	output
;		Ȃ
;	break
;		a, b, c, d, e, f, h, l
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_mute_psg::
		di
		xor		a, a
		ld		d, a
		ld		c, PSG_REG_ADR
		ld		b, 16

bgmdriver_mute_psg_loop:
		out		[c], d					; FOR b=0 TO 15:SOUND b, 0:NEXT
		inc		d
		out		[PSG_REG_WRT], a
		djnz	bgmdriver_mute_psg_loop

		ld		d, 7					; SOUND 7, &H80 + &H3F
		out		[c], d
		ld		a, 0x80 + 0x3F
		out		[PSG_REG_WRT], a
		ei

		ret

; -----------------------------------------------------------------------------
;	ʉJn
;	input:
;		hl	...	ʉf[^̃AhX
;	output
;		Ȃ
;	break
;		a
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_play_sound_effect::
		; ʉJn̂߂̏{
		; page1  RAM ɐ؂ւ
		push	hl
		di
		ld		a, [play_sound_effect_priority]			; Đ̌ʉvCIeBH
		cp		a, [hl]
		jp		c, bgmdriver_play_sound_effect_skip		; Ⴏ΍ĐȂ
		ld		a, [hl]
		ld		[play_sound_effect_priority], a			; vCIeBXV
		inc		hl
		ld		[play_sound_effect_adr], hl				; ʉf[^̃AhX
		xor		a, a
		ld		[play_sound_effect_wait_count], a		; ҋ@ 0
		ld		[play_sound_effect_freq+0], a			; Đg 0
		ld		[play_sound_effect_freq+1], a
		ld		[play_sound_effect_noise_freq], a		; mCYg 0
		ld		[play_sound_effect_volume], a			;  0
		inc		a
		ld		[play_sound_effect_active], a			; ʉĐJn
bgmdriver_play_sound_effect_skip:
		pop		hl
		ei
		ret

; -----------------------------------------------------------------------------
;	t^XN[]
;	input:
;		hl	...	鉉t^XÑAhX
;	output
;		Ȃ
;	break
;		a, b, c, d, e, h, l
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_init_play_info:
		ld		e, l
		ld		d, h
		inc		de
		ld		bc, INFO_SIZE - 1
		xor		a, a
		ld		[hl], a
		ldir
		ret

; -----------------------------------------------------------------------------
;	t[`
;	input:
;		Ȃ
;	output
;		Ȃ
;	break
;		Ȃ
;	comment
;		1/60bԊu call 邱Ƃ҂Ă郋[`łAʏ
;		H_TIMI tbN[`ĂяoB
;		荞ݏz肵Ă邽߁AWX^͔j󂵂ȂB
; -----------------------------------------------------------------------------
bgmdriver_interrupt_handler::
		; t[`Ăяo
		ld		ix, play_info_ch0
		call	bgmdriver_play_ch
		ld		ix, play_info_ch1
		call	bgmdriver_play_ch
		ld		ix, play_info_ch2
		call	bgmdriver_play_ch
		; ʉ[`Ăяo
		call	bgmdriver_sound_effect
		; tF[hAEg
		call	bgmdriver_fadeout_proc
		; ~LT[[`Ăяo
		call	bgmdriver_mixer
		ret

; -----------------------------------------------------------------------------
;	t[`[1ch] []
;	input:
;		ix	...	t{鉉t^XÑAhX
;	output
;		Ȃ
;	break
;		a, b, c, d, e, f, h, l
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_play_ch:
		; tł邩ׂ
		ld		a, [ix + INFO_PLAY_ADR_L]
		or		a, [ix + INFO_PLAY_ADR_H]
		ret		z									; tłȂΉɒE
		; ҋ@Ԃł邩ׂ
		ld		l, [ix + INFO_WAIT_COUNT_L]
		ld		h, [ix + INFO_WAIT_COUNT_H]
		ld		a, l
		or		a, h
		jr		z, bgmdriver_check_next_data
		; ҋ@ԂXV
		dec		hl
		ld		[ix + INFO_WAIT_COUNT_L], l
		ld		[ix + INFO_WAIT_COUNT_H], h
		jp		bgmdriver_update_vibrato
		; ̉tf[^ǂݎ
bgmdriver_check_next_data:
		ld		l, [ix + INFO_PLAY_ADR_L]
		ld		h, [ix + INFO_PLAY_ADR_H]
		ld		a, [hl]
		inc		hl
		; tf[^fR[h
		cp		a, 96
		jp		c, bgmdriver_keyon					; 0`95  KeyOn bZ[W
		cp		a, 101
		jp		c, bgmdriver_drum_keyon				; 96`100  KeyOn bZ[W
		jp		z, bgmdriver_keyoff					; 101  KeyOff bZ[W
		cp		a, 103
		jp		c, bgmdriver_rest					; 102  xbZ[W
		jp		z, bgmdriver_volume					; 103  ʐݒ胁bZ[W
		cp		a, 105
		jp		c, bgmdriver_sound_font				; 104  Fݒ胁bZ[W
		jp		z, bgmdriver_jump					; 105  AhXWvbZ[W
		cp		a, 107
		jp		c, bgmdriver_play_end				; 106  t~bZ[W
		jp		z, bgmdriver_drum1_font				; 107  hPFݒ胁bZ[W
		cp		a, 109
		jp		c, bgmdriver_drum2_font				; 108  hQFݒ胁bZ[W
		jp		z, bgmdriver_drum3_font				; 109  hRFݒ胁bZ[W
		cp		a, 111
		jp		c, bgmdriver_drum4_font				; 110  hSFݒ胁bZ[W
		jp		z, bgmdriver_drum5_font				; 111  hTFݒ胁bZ[W
		ret

		; KeyOn -----------------------------------------------------------
bgmdriver_keyon:
		ld		c, a								; j󂳂ȂWX^ɃobNAbv
		call	bgmdriver_get_wait_time
		ld		[ix + INFO_WAIT_COUNT_L], e			; ҂ԍXV
		ld		[ix + INFO_WAIT_COUNT_H], d
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		a, c								; 
		; Kgɕϊ
		rlca
		ld		l, a
		ld		h, 0
		ld		de, freq_data
		add		hl, de								; hl  freq_data + a * 2
		ld		e, [hl]
		inc		hl
		ld		d, [hl]
		; Ff[^擾
		ld		l, [ix + INFO_SEL_SFONT_L]
		ld		h, [ix + INFO_SEL_SFONT_H]
		; ݒgXV
bgmdriver_set_freq:
		ld		[ix + INFO_TONE_FREQ_L], e
		ld		[ix + INFO_TONE_FREQ_H], d
		; e평
		xor		a, a
		ld		[ix + INFO_ENV_STATE], a
		ld		[ix + INFO_ENV_VOL], a
		; ru[g
		ld		[ix + INFO_VIB_INDEX], a
		ld		[ix + INFO_SOUND_FONT_L], l
		ld		[ix + INFO_SOUND_FONT_H], h
		ld		de, SFONT_VIB_WAIT
		add		hl, de
		ld		a, [hl]
		ld		[ix + INFO_VIB_WAIT], a
		; mCYg
		inc		hl
		ld		a, [hl]
		ld		[ix + INFO_NOISE_FREQ], a
		jp		bgmdriver_update_vibrato

		; DRUM KeyOn -------------------------------------------------------
bgmdriver_drum_keyon:
		ld		c, a								; j󂳂ȂWX^ɃobNAbv
		call	bgmdriver_get_wait_time
		ld		[ix + INFO_WAIT_COUNT_L], e			; ҂ԍXV
		ld		[ix + INFO_WAIT_COUNT_H], d
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		a, c								; 
		; F擾
		sub		a, 96								; hl  [a - 96] * 2 + play_drum_font1
		rlca
		ld		hl, play_drum_font1
		add		a, l
		ld		l, a								; tOs
		ld		a, 0								; tOs
		adc		a, h
		ld		h, a								; hl ɃhF̃AhX̓ĂAhX
		ld		e, [hl]
		inc		hl
		ld		d, [hl]
		ex		de, hl								; hl ɃhF̃AhX
		push	hl
		ld		de, SFONT_FREQ_L
		add		hl, de
		ld		e, [hl]
		inc		hl
		ld		d, [hl]								; de ɃhpĐg
		pop		hl									; hl ɃhF̃AhX
		jp		bgmdriver_set_freq

		; KeyOff -----------------------------------------------------------
bgmdriver_keyoff:
		call	bgmdriver_get_wait_time
		ld		[ix + INFO_WAIT_COUNT_L], e			; ҂ԍXV
		ld		[ix + INFO_WAIT_COUNT_H], d
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		; Gx[v[XԂɕύX
		ld		a, [ix + INFO_ENV_STATE]
		cp		a, 3
		jp		nc, bgmdriver_update_vibrato
		ld		[ix + INFO_ENV_STATE], 3
		jp		bgmdriver_update_vibrato

		; x̏ ----------------------------------------------------------
bgmdriver_rest:
		call	bgmdriver_get_wait_time
		ld		[ix + INFO_WAIT_COUNT_L], e			; ҂ԍXV
		ld		[ix + INFO_WAIT_COUNT_H], d
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		[ix + INFO_ENV_STATE], 4
		ld		[ix + INFO_ENV_VOL], 0
		jp		bgmdriver_update_vibrato

		; ʐݒ ------------------------------------------------------------
bgmdriver_volume:
		ld		a, [hl]								; ʎ擾
		inc		hl
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		[ix + INFO_TONE_VOL], a				; ݒ艹ʍXV
		jp		bgmdriver_check_next_data			; Ď̃f[^

		; Fݒ ------------------------------------------------------------
bgmdriver_sound_font:
		ld		e, [hl]								; Ff[^̃AhX擾
		inc		hl
		ld		d, [hl]
		inc		hl
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		hl, [play_bgm_data_adr]				; tf[^̐擪AhX
		add		hl, de
		ld		[ix + INFO_SEL_SFONT_L], l				; Ff[^AhXXV
		ld		[ix + INFO_SEL_SFONT_H], h
		xor		a, a									; e평
		ld		[ix + INFO_VIB_WAIT], a
		ld		[ix + INFO_ENV_STATE], a
		ld		[ix + INFO_ENV_VOL], a
		ld		[ix + INFO_VIB_INDEX], a
		jp		bgmdriver_check_next_data			; Ď̃f[^

		; hPFݒݒ ------------------------------------------------
bgmdriver_drum1_font:
		ld		e, [hl]								; Ff[^̃AhX擾
		inc		hl
		ld		d, [hl]
		inc		hl
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		hl, [play_bgm_data_adr]				; tf[^̐擪AhX
		add		hl, de
		ld		[play_drum_font1], hl				; Ff[^AhXXV
		xor		a, a									; e평
		ld		[ix + INFO_VIB_WAIT], a
		ld		[ix + INFO_ENV_STATE], a
		ld		[ix + INFO_ENV_VOL], a
		ld		[ix + INFO_VIB_INDEX], a
		jp		bgmdriver_check_next_data			; Ď̃f[^

		; hQFݒݒ ------------------------------------------------
bgmdriver_drum2_font:
		ld		e, [hl]								; Ff[^̃AhX擾
		inc		hl
		ld		d, [hl]
		inc		hl
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		hl, [play_bgm_data_adr]				; tf[^̐擪AhX
		add		hl, de
		ld		[play_drum_font2], hl				; Ff[^AhXXV
		xor		a, a									; e평
		ld		[ix + INFO_VIB_WAIT], a
		ld		[ix + INFO_ENV_STATE], a
		ld		[ix + INFO_ENV_VOL], a
		ld		[ix + INFO_VIB_INDEX], a
		jp		bgmdriver_check_next_data			; Ď̃f[^

		; hRFݒݒ ------------------------------------------------
bgmdriver_drum3_font:
		ld		e, [hl]								; Ff[^̃AhX擾
		inc		hl
		ld		d, [hl]
		inc		hl
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		hl, [play_bgm_data_adr]				; tf[^̐擪AhX
		add		hl, de
		ld		[play_drum_font3], hl				; Ff[^AhXXV
		xor		a, a									; e평
		ld		[ix + INFO_VIB_WAIT], a
		ld		[ix + INFO_ENV_STATE], a
		ld		[ix + INFO_ENV_VOL], a
		ld		[ix + INFO_VIB_INDEX], a
		jp		bgmdriver_check_next_data			; Ď̃f[^

		; hSFݒݒ ------------------------------------------------
bgmdriver_drum4_font:
		ld		e, [hl]								; Ff[^̃AhX擾
		inc		hl
		ld		d, [hl]
		inc		hl
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		hl, [play_bgm_data_adr]				; tf[^̐擪AhX
		add		hl, de
		ld		[play_drum_font4], hl				; Ff[^AhXXV
		xor		a, a									; e평
		ld		[ix + INFO_VIB_WAIT], a
		ld		[ix + INFO_ENV_STATE], a
		ld		[ix + INFO_ENV_VOL], a
		ld		[ix + INFO_VIB_INDEX], a
		jp		bgmdriver_check_next_data			; Ď̃f[^

		; hTFݒݒ ------------------------------------------------
bgmdriver_drum5_font:
		ld		e, [hl]								; Ff[^̃AhX擾
		inc		hl
		ld		d, [hl]
		inc		hl
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		ld		hl, [play_bgm_data_adr]				; tf[^̐擪AhX
		add		hl, de
		ld		[play_drum_font5], hl				; Ff[^AhXXV
		xor		a, a									; e평
		ld		[ix + INFO_VIB_WAIT], a
		ld		[ix + INFO_ENV_STATE], a
		ld		[ix + INFO_ENV_VOL], a
		ld		[ix + INFO_VIB_INDEX], a
		jp		bgmdriver_check_next_data			; Ď̃f[^

		; AhXWv ----------------------------------------------------
bgmdriver_jump:
		ld		e, [hl]								; ѐAhX擾
		inc		hl
		ld		d, [hl]
		ld		hl, [play_bgm_data_adr]				; tf[^̐擪AhX
		add		hl, de
		ld		[ix + INFO_PLAY_ADR_L], l				; tf[^
		ld		[ix + INFO_PLAY_ADR_H], h
		jp		bgmdriver_check_next_data			; Ď̃f[^

		; t~ ------------------------------------------------------------
bgmdriver_play_end:
		xor		a, a
		ld		[ix + INFO_PLAY_ADR_L], a				; t~
		ld		[ix + INFO_PLAY_ADR_H], a
		ld		[ix + INFO_TONE_VOL], a
		ld		[ix + INFO_ENV_STATE], a
		ld		[ix + INFO_ENV_VOL], a
		jp		bgmdriver_envelope_end				; ~̂Ńru[gGx[v̓XLbv

		; ru[g̏ ----------------------------------------------------
bgmdriver_update_vibrato:
		; ru[gxԂ̍Œł邩ׂ
		ld		a, [ix + INFO_VIB_WAIT]
		or		a, a
		jr		z, bgmdriver_vibrato_active
		; ru[gxԂ̍Œ̏ꍇ
		dec		a
		ld		[ix + INFO_VIB_WAIT], a				; x1JEgo
		; ݒg擾
		ld		l, [ix + INFO_TONE_FREQ_L]
		ld		h, [ix + INFO_TONE_FREQ_H]
		jp		bgmdriver_update_freq
		; ru[gxԂEĂꍇ
bgmdriver_vibrato_active:
		; ru[gʑʒu擾
		ld		a, [ix + INFO_VIB_INDEX]
		ld		e, a
		inc		a
		and		a, 31
		ld		[ix + INFO_VIB_INDEX], a				; ̈ʑ֐i߂Ă
		; ʑʒuɑΉru[gg`擾
		ld		d, 0
		ld		l, [ix + INFO_SOUND_FONT_L]
		ld		h, [ix + INFO_SOUND_FONT_H]
		add		hl, de
		ld		a, [hl]								; a  INFO_SOUND_FONT->SFONT_VIB_WAVE[ INFO_VIB_INDEX ]
		; ݒg擾
		ld		l, [ix + INFO_TONE_FREQ_L]
		ld		h, [ix + INFO_TONE_FREQ_H]
		; ru[gg`Z
		or		a, a
		jp		p, bgmdriver_vibrato_active_skip
		ld		d, 255
bgmdriver_vibrato_active_skip:
		ld		e, a
		add		hl, de
		; PSGɐݒ肳gXV
bgmdriver_update_freq:
		ld		[ix + INFO_EFF_FREQ_L], l
		ld		[ix + INFO_EFF_FREQ_H], h

		; Gx[v̏ --------------------------------------------------
		ld		l, [ix + INFO_SOUND_FONT_L]
		ld		h, [ix + INFO_SOUND_FONT_H]
		ld		d, 0
		ld		a, [ix + INFO_ENV_STATE]				; 0: AR, 1: DR, 2: SR, 3: RR, 4: ~
		sub		a, 1
		jr		c, bgmdriver_envelope_ar
		jr		z, bgmdriver_envelope_dr
		sub		a, 2
		jr		c, bgmdriver_envelope_sr
		jr		z, bgmdriver_envelope_rr
		jp		bgmdriver_envelope_end
		; [XCg̏
bgmdriver_envelope_rr:
		ld		a, [ix + INFO_ENV_VOL]
		ld		e, SFONT_RR
		add		hl, de
		ld		b, [hl]
		sub		a, b
		ld		[ix + INFO_ENV_VOL], a					; tOs
		jr		nc, bgmdriver_envelope_end
		ld		[ix + INFO_ENV_VOL], 0
		ld		[ix + INFO_ENV_STATE], 4				; ~Ɉڍs
		jp		bgmdriver_envelope_end
		; TXeBCg̏
bgmdriver_envelope_sr:
		ld		a, [ix + INFO_ENV_VOL]
		ld		e, SFONT_SR
		add		hl, de
		ld		b, [hl]
		sub		a, b
		ld		[ix + INFO_ENV_VOL], a					; tOs
		jr		nc, bgmdriver_envelope_end
		ld		[ix + INFO_ENV_VOL], 0
		ld		[ix + INFO_ENV_STATE], 4				; ~Ɉڍs
		jp		bgmdriver_envelope_end
		; fBPCCg̏
bgmdriver_envelope_dr:
		ld		a, [ix + INFO_ENV_VOL]
		ld		e, SFONT_DR
		add		hl, de
		ld		b, [hl]								; SFONT_DR
		inc		hl
		ld		c, [hl]								; SFONT_SL
		sub		a, b
		ld		[ix + INFO_ENV_VOL], a					; tOs
		cp		a, c
		jr		nc, bgmdriver_envelope_end
		ld		[ix + INFO_ENV_VOL], c
		ld		[ix + INFO_ENV_STATE], 2				; SR Ɉڍs
		jp		bgmdriver_envelope_end
		; A^bNCg̏
bgmdriver_envelope_ar:
		ld		a, [ix + INFO_ENV_VOL]
		ld		e, SFONT_AR
		add		hl, de
		ld		b, [hl]
		add		a, b
		ld		[ix + INFO_ENV_VOL], a					; tOs
		jr		nc, bgmdriver_envelope_end
		ld		[ix + INFO_ENV_VOL], 255
		ld		[ix + INFO_ENV_STATE], 1				; DR Ɉڍs
		jp		bgmdriver_envelope_end
bgmdriver_envelope_end:

		; ʂvZ ------------------------------------------------------
		ld		a, [ix + INFO_TONE_VOL]
		xor		a, 15
		ld		b, a
		ld		a, [ix + INFO_ENV_VOL]
		srl		a
		srl		a
		srl		a
		srl		a
		sub		a, b
		jr		nc, bgmdriver_calc_eff_vol
		xor		a, a
bgmdriver_calc_eff_vol:
		ld		[ix + INFO_EFF_VOL], a
		ret

; -----------------------------------------------------------------------------
;	҂ԓǂݎ菈 []
;	input:
;		hl	...	҂ԂL^Ă郁̃AhX
;	output
;		hl	...	҂Ԃ̎̃AhX
;		de	...	ǂݎ҂
;	break
;		a, f, d, e, h, l
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_get_wait_time:
		ld		de, 0
bgmdriver_get_wait_time_loop:
		ld		a, [hl]
		inc		hl
		inc		a
		jr		nz, bgmdriver_get_wait_time_exit
		dec		a
		add		a, e
		jr		nc, bgmdriver_get_wait_time_loop
		inc		d
		jr		bgmdriver_get_wait_time_loop
bgmdriver_get_wait_time_exit:
		dec		a
		add		a, e
		ld		e, a
		jr		nc, bgmdriver_get_wait_time_skip
		inc		d
bgmdriver_get_wait_time_skip:
		ret

; -----------------------------------------------------------------------------
;	tF[hAEg []
;	input:
;		Ȃ
;	output
;		Ȃ
;	break
;		
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_fadeout_proc:
		; tF[hAEg쒆f
		ld		a, [play_master_volume_speed]
		or		a, a
		ret		z

		; ҋ@f
		ld		a, [play_master_volume_wait]
		or		a, a
		jr		z, bgmdriver_fadeout_skip1
		; ҋ@Ȃ҂Ԍ
		dec		a
		ld		[play_master_volume_wait], a
		ret
		; ҋ@łȂ
bgmdriver_fadeout_skip1:
		; ̑҂Ԃݒ
		ld		a, [play_master_volume_speed]
		ld		[play_master_volume_wait], a
		; ʌ
		ld		a, [play_master_volume]
		inc		a								; 0ő剹, 15Ȃ̂ŁAinc a ŉʌ
		ld		[play_master_volume], a
		cp		a, 15
		ret		nz
		; ɂȂtF[hAEg~
		xor		a, a
		ld		[play_master_volume_speed], a
		; }X^[ʂőɖ߂
		ld		[play_master_volume], a
		; t~
		ld		hl, play_info_ch0
		call	bgmdriver_init_play_info
		ld		hl, play_info_ch1
		call	bgmdriver_init_play_info
		ld		hl, play_info_ch2
		call	bgmdriver_init_play_info
		ret

; -----------------------------------------------------------------------------
;	ʉ []
;	input:
;		Ȃ
;	output
;		Ȃ
;	break
;		
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_sound_effect:
		; ʉĐf
		ld		a, [play_sound_effect_active]
		or		a, a
		ret		z

		; ҋ@Ԓf
		ld		a, [play_sound_effect_wait_count]
		or		a, a
		jr		z, bgmdriver_sound_effect_proc
		; ҋ@Ԍo
		dec		a
		ld		[play_sound_effect_wait_count], a
		ret

		; ʉ
bgmdriver_sound_effect_proc:
		ld		hl, [play_sound_effect_adr]
bgmdriver_sound_effect_loop:
		ld		a, [hl]
		inc		hl
		ld		[play_sound_effect_adr], hl				; ̃AhX

		; R[h̉
		cp		a, BGM_SE_VOL
		jp		c, bgmdriver_sound_effect_freq_proc
		jp		z, bgmdriver_sound_effect_volume_proc
		cp		a, BGM_SE_WAIT
		jp		c, bgmdriver_sound_effect_noise_freq_proc
		jp		z, bgmdriver_sound_effect_wait_proc
		jp		bgmdriver_sound_effect_end_proc

		; ʉ̎gݒ
bgmdriver_sound_effect_freq_proc:
		ld		e, [hl]
		inc		hl
		ld		d, [hl]
		inc		hl
		ld		[play_sound_effect_adr], hl				; ̃AhX
		ld		[play_sound_effect_freq], de			; gXV
		jp		bgmdriver_sound_effect_loop

		; ʉ̉ʐݒ
bgmdriver_sound_effect_volume_proc:
		ld		a, [hl]									; ʂ擾
		inc		hl
		ld		[play_sound_effect_volume], a			; ʂXV
		jp		bgmdriver_sound_effect_loop

		; ʉ̃mCYgݒ
bgmdriver_sound_effect_noise_freq_proc:
		ld		a, [hl]									; mCYg擾
		inc		hl
		ld		[play_sound_effect_noise_freq], a		; mCYgXV
		jp		bgmdriver_sound_effect_loop

		; Pҋ@
bgmdriver_sound_effect_wait_proc:
		ld		a, [hl]									; ҋ@Ԃ擾
		inc		hl
		ld		[play_sound_effect_adr], hl				; ̃AhX
		ld		[play_sound_effect_wait_count], a		; ҋ@ԂXV
		ret

		; ʉ~
bgmdriver_sound_effect_end_proc:
		xor		a, a
		ld		[play_sound_effect_active], a
		dec		a
		ld		[play_sound_effect_priority], a			; ŒvCIeBɍXV
		ret

; -----------------------------------------------------------------------------
;	~LT[ []
;	input:
;		Ȃ
;	output
;		Ȃ
;	break
;		a, b, c, d, e, f, h, l, ix
;	comment
;		Ȃ
; -----------------------------------------------------------------------------
bgmdriver_mixer:
		ld		b, 0							; tone  on/off tO [1  on]
		ld		e, 0							; noise  on/off tO [1  on]
		ld		c, PSG_REG_ADR

		; ch0 gݒ
		ld		d, 0
		ld		ix, play_info_ch0
		ld		a, [ix + INFO_EFF_FREQ_L]			; SOUND 0, [ix + INFO_EFF_FREQ_L]
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a
		ld		a, [ix + INFO_EFF_FREQ_H]			; SOUND 1, [ix + INFO_EFF_FREQ_H]
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a
		; ch0 ̎ޔ
		and		a, 0x80
		jr		nz, bgmdriver_mixer_skip1_ch0	; g[off ȂXLbv
		inc		b								; ch0 ̃g[on ێ
bgmdriver_mixer_skip1_ch0:
		ld		a, [ix + INFO_NOISE_FREQ]
		bit		7, a
		jr		z, bgmdriver_mixer_skip2_ch0	; mCYoff ȂXLbv
		and		a, 31
		ld		[play_noise_freq], a			; mCYgoĂ
		ld		e, 8							; ch0 ̃mCYon ێ
bgmdriver_mixer_skip2_ch0:

		; ch1 gݒ
		ld		ix, play_info_ch1
		ld		a, [ix + INFO_EFF_FREQ_L]			; SOUND 2, [ix + INFO_EFF_FREQ_L]
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a
		ld		a, [ix + INFO_EFF_FREQ_H]			; SOUND 3, [ix + INFO_EFF_FREQ_H]
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a
		; ch1 ̎ޔ
		and		a, 0x80
		jr		nz, bgmdriver_mixer_skip1_ch1	; g[off ȂXLbv
		inc		b								; ch0 ̃g[on ێ
		inc		b
bgmdriver_mixer_skip1_ch1:
		ld		a, [ix + INFO_NOISE_FREQ]
		bit		7, a
		jr		z, bgmdriver_mixer_skip2_ch1	; mCYoff ȂXLbv
		and		a, 31
		ld		[play_noise_freq], a			; mCYgoĂ
		ld		a, 16							; ch1 ̃mCYon ێ
		add		a, e
		ld		e, a
bgmdriver_mixer_skip2_ch1:

		; ch2  BGM ʉ
		ld		a, [play_sound_effect_active]
		or		a, a
		jr		z, bgmdriver_mixer_tone_ch2

		; ch2 ʉ̎gݒ
		ld		hl, [play_sound_effect_freq]
		ld		a, l							; SOUND 4, l
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a
		ld		a, h							; SOUND 5, h
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a
		; ch2 ʉ̃g[
		and		a, 0x80
		jr		nz, bgmdriver_mixer_skip0_ch2	; g[off ȂXLbv
		inc		b
		inc		b
		inc		b
		inc		b

bgmdriver_mixer_skip0_ch2:
		; ch2 ʉ̃mCYgݒ
		ld		a, [play_sound_effect_noise_freq]
		bit		7, a
		jp		z, bgmdriver_mixer_noise_freq
		and		a, 0x3F
		out		[c], d							; SOUND 6, play_sound_effect_noise_freq
		inc		d
		out		[PSG_REG_WRT], a
		ld		a, 32							; ch1 ̃mCYon ێ
		add		a, e
		ld		e, a
		jp		bgmdriver_mixer_mix

bgmdriver_mixer_tone_ch2:
		; ch2 gݒ
		ld		ix, play_info_ch2
		ld		a, [ix + INFO_EFF_FREQ_L]		; SOUND 4, [ix + INFO_EFF_FREQ_L]
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a
		ld		a, [ix + INFO_EFF_FREQ_H]		; SOUND 5, [ix + INFO_EFF_FREQ_H]
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a
		; ch2 ̎ޔ
		and		a, 0x80
		jr		nz, bgmdriver_mixer_skip1_ch2	; g[off ȂXLbv
		inc		b
		inc		b
		inc		b
		inc		b
bgmdriver_mixer_skip1_ch2:
		ld		a, [ix + INFO_NOISE_FREQ]
		bit		7, a
		jr		z, bgmdriver_mixer_skip2_ch2	; mCYoff ȂXLbv
		and		a, 31
		ld		[play_noise_freq], a			; mCYgoĂ
		ld		a, 32							; ch1 ̃mCYon ێ
		add		a, e
		ld		e, a
bgmdriver_mixer_skip2_ch2:

bgmdriver_mixer_noise_freq:
		; mCYg
		out		[c], d							; SOUND 6, play_noise_freq
		inc		d
		ld		a, [play_noise_freq]
		out		[PSG_REG_WRT], a

bgmdriver_mixer_mix:
		; ~LT[
		ld		hl, play_master_volume
		out		[c], d							; SOUND 7, [b | e | 0x80] ^ 0x3F
		inc		d
		ld		a, e
		or		a, b
		or		a, 0x80
		xor		a, 0x3F
		out		[PSG_REG_WRT], a

		; ch0 ʐݒ
		ld		ix, play_info_ch0				; SOUND 8, [ix + INFO_EFF_VOL]
		ld		a, [ix + INFO_EFF_VOL]
		sub		a, [hl]
		jp		nc, bgmdriver_mixer_mix_skip1
		xor		a, a
bgmdriver_mixer_mix_skip1:
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a

		; ch1 ʐݒ
		ld		ix, play_info_ch1				; SOUND 9, [ix + INFO_EFF_VOL]
		ld		a, [ix + INFO_EFF_VOL]
		sub		a, [hl]
		jp		nc, bgmdriver_mixer_mix_skip2
		xor		a, a
bgmdriver_mixer_mix_skip2:
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a

		; ch2  BGM ʉ
		ld		a, [play_sound_effect_active]
		or		a, a
		jr		z, bgmdriver_mixer_volume_ch2
		ld		a, [play_sound_effect_volume]
		jp		bgmdriver_mixer_volume_ch2_skip1
bgmdriver_mixer_volume_ch2:
		; ch2 ʐݒ
		ld		ix, play_info_ch2				; SOUND 10, [ix + INFO_EFF_VOL]
		ld		a, [ix + INFO_EFF_VOL]
		sub		a, [hl]
		jp		nc, bgmdriver_mixer_mix_skip3
		xor		a, a
bgmdriver_mixer_mix_skip3:

bgmdriver_mixer_volume_ch2_skip1:
		out		[c], d
		inc		d
		out		[PSG_REG_WRT], a

		ret

; -----------------------------------------------------------------------------
;	f[^GA
; -----------------------------------------------------------------------------
freq_data:
		dw		3420, 3228, 3047, 2876, 2714, 2562, 2418, 2282, 2154, 2033, 1919, 1811
		dw		1710, 1614, 1523, 1438, 1357, 1281, 1209, 1141, 1077, 1016,  959,  905
		dw		 855,  807,  761,  719,  678,  640,  604,  570,  538,  508,  479,  452
		dw		 427,  403,  380,  359,  339,  320,  302,  285,  269,  254,  239,  226
		dw		 213,  201,  190,  179,  169,  160,  151,  142,  134,  127,  119,  113
		dw		 106,  100,   95,   89,   84,   80,   75,   71,   67,   63,   59,   56
		dw		  53,   50,   47,   44,   42,   40,   37,   35,   33,   31,   29,   28
		dw		  26,   25,   23,   22,   21,   20,   18,   17,   16,   15,   14,   14

; -----------------------------------------------------------------------------
;	[NGA
; -----------------------------------------------------------------------------
play_sound_effect_active:
		db		0				; ʉĐ 1
play_sound_effect_wait_count:
		db		0				; ʉ̑ҋ@
play_sound_effect_freq:
		dw		0				; ʉ̍Đg
play_sound_effect_noise_freq:
		db		0				; ʉ̃mCYg
play_sound_effect_volume:
		db		0				; ʉ̉
play_sound_effect_adr:
		dw		0				; Đ̌ʉf[^̃AhX
play_sound_effect_priority:
		db		255				; Đ̌ʉ̃vCIeB [0ō]

play_noise_freq:
		db		0				; ۂɍĐmCYgpƕϐ

play_bgm_data_adr:
		dw		0				; Đ BGMf[^擪AhX

play_master_volume_wait:
		db		0				; tF[hAEgpҋ@
play_master_volume_speed:
		db		0				; tF[hAEgpҋ@ԏl[0̓tF[hAEg~]
play_master_volume:
		db		0				; }X^[[0ő剹, 15]

play_drum_font1:
		dw		0				; hP̉Ff[^AhX
play_drum_font2:
		dw		0				; hQ̉Ff[^AhX
play_drum_font3:
		dw		0				; hR̉Ff[^AhX
play_drum_font4:
		dw		0				; hS̉Ff[^AhX
play_drum_font5:
		dw		0				; hT̉Ff[^AhX

play_info_ch0:
		repeat i, INFO_SIZE
			db		0			; ch0 ̉tf[^
		endr
play_info_ch1:
		repeat i, INFO_SIZE
			db		0			; ch1 ̉tf[^
		endr
play_info_ch2:
		repeat i, INFO_SIZE
			db		0			; ch2 ̉tf[^
		endr
